//#define TESTING
//#define COLORTEST
#define JOYSTICKS
//#define LOGLAST

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;

using System.Drawing;
using System.Drawing.Imaging;

using OpenTK;
using OpenTK.Input;
using OpenTK.Graphics.OpenGL;

using OpenTK.Audio;

namespace OpenTKTest
{
    /// <summary>
    /// Demonstrates simple OpenGL Texturing.
    /// </summary>
    public class Textures : GameWindow
    {
        GLTexture2D texSquareGuy;

        const int xMin = -1;
        const int xMax = 18;
        const int fMin = -1;
        const int fMax = 12;

        readonly int objs = (xMax - xMin) * (fMax - fMin);
        
#if LOGLAST
        private void LogLast(string last)
        {
            last = "The last thing I did was: " + last + ".";
            System.Diagnostics.Debug.WriteLine(last);
            File.WriteAllText("last.log", last);
        }
#endif

        private string baseTitle
        {
            get
            {
                string ret = "OpenTK Texture Test (" + objs + " objects)";

                return ret;
            }
        }

        public Textures()
            : base(800, 600, OpenTK.Graphics.GraphicsMode.Default)
        {
            //this.WindowBorder = OpenTK.WindowBorder.Hidden;
        }
        private const double RotSteps = 1;
		double RotAmount;

        #region OnLoad

        /// <summary>
        /// Setup OpenGL and load resources here.
        /// </summary>
        /// <param name="e">Not used.</param>
        protected override void OnLoad(EventArgs e)
        {
            this.Title = baseTitle;
            GL.ClearColor(Color.MidnightBlue);
            GL.Enable(EnableCap.Texture2D);
            GL.Enable(EnableCap.SampleAlphaToCoverage);
            GL.Hint(HintTarget.PerspectiveCorrectionHint, HintMode.Nicest);
            texSquareGuy = new GLTexture2D("SquareGuy.png");

            RotAmount = -RotSteps;

            AudioRun();

        }

        #endregion

        #region OnUnload

        protected override void OnUnload(EventArgs e)
        {
            texSquareGuy.Dispose();

            AudioEnd();

#if LOGLAST
            LogLast("successfully unload the program");
#endif
        }

        #endregion

        #region OnResize

        /// <summary>
        /// Respond to resize events here.
        /// </summary>
        /// <param name="e">Contains information on the new GameWindow size.</param>
        /// <remarks>There is no need to call the base implementation.</remarks>
        protected override void OnResize(EventArgs e)
        {
            GL.Viewport(0, 0, Width, Height);

            GL.MatrixMode(MatrixMode.Projection);
            GL.LoadIdentity();
            GL.Ortho(-1.0, 1.0, -1.0, 1.0, 0.0, 4.0);
        }

        #endregion

        #region Begin2D
        public void Begin2D()
        {
            GL.MatrixMode(MatrixMode.Projection);

            GL.LoadIdentity();

            GL.Ortho(0, Width, Height, 0, 0, 1);
            
            GL.Disable(EnableCap.DepthTest);

            GL.MatrixMode(MatrixMode.Modelview);

            GL.LoadIdentity();

            //GL.Translate(0.375, 0.375, 0);

            GL.Clear(ClearBufferMask.ColorBufferBit);

            GL.Enable(EnableCap.Blend);
        }
        #endregion

        #region OnUpdateFrame

        double UFPS = 0;
        double GFPS = 0;
        double GDFPS = 0;
        int UFrameCount = 0;
        int GFrameCount = 0;
        bool Focus = true;
        /// <summary>
        /// Add your game logic here.
        /// </summary>
        /// <param name="e">Contains timing information.</param>
        /// <remarks>There is no need to call the base implementation.</remarks>
        protected override void OnUpdateFrame(FrameEventArgs e)
        {
            if (!this.Focused)
            {
                if(!Focus) return;
                Focus = false;
                if (!audioStopped) AL.SourcePause(SoundSource);
            }
            else if (!Focus)
            {
                Focus = true;
                if (!audioStopped) AL.SourcePlay(SoundSource);
            }

            if (Keyboard[Key.Escape] || 
#if JOYSTICKS
                (Joysticks.Count > 0 && Joysticks[0].Button[JoystickButton.Button1]) ||
#endif
                (Keyboard[Key.F4] && (Keyboard[Key.LAlt] || Keyboard[Key.RAlt])))
            {
                this.Exit();
            }


            #region Stop sound!
            if (Keyboard[Key.Enter] 
#if JOYSTICKS
                || (Joysticks.Count > 0 && Joysticks[0].Button[0])
#endif
                )
            {
                if (!enterPressed)
                {
                    enterPressed = true;

                    if (audioStopped)
                    {
                        AL.SourcePlay(SoundSource);
                        audioStopped = false;
                    }
                    else
                    {
                        AL.SourcePause(SoundSource);
                        audioStopped = true;
                    }
                }
            }
            else enterPressed = false;
            #endregion

            #region Changing of The Audio

            int getProcessed;
            int getQueued;
            AL.GetSource(SoundSource, ALGetSourcei.BuffersProcessed,
                out getProcessed);
            AL.GetSource(SoundSource, ALGetSourcei.BuffersQueued,
                out getQueued);
            if (getQueued - getProcessed == 1)
            {
                AL.SourceQueueBuffer(SoundSource, SoundBuffer[1]);
            }
            #endregion


            UFPS += e.Time;
            if (UFrameCount == (int)Math.Round(1 / this.TargetUpdatePeriod))
            {
                UFrameCount = 0;
                this.Title = baseTitle +
                    " / FPS:  " + Math.Round(UFPS * 60, 2)
                    + " updates, " + Math.Round(GDFPS * 60, 2) + " renders";
                UFPS = 0;
            }
            UFrameCount++;
            RotAmount = (RotAmount + RotSteps) % 360;
        }

        #endregion

        #region OnRenderFrame

        /// <summary>
        /// Add your game rendering code here.
        /// </summary>
        /// <param name="e">Contains timing information.</param>
        /// <remarks>There is no need to call the base implementation.</remarks>
        protected override void OnRenderFrame(FrameEventArgs e)
        {
            GFPS += e.Time;
            if (GFrameCount == (int)Math.Round(1 / this.TargetUpdatePeriod))
            {
                GFrameCount = 0;
                GDFPS = GFPS;
                GFPS = 0;
            }
            GFrameCount++;

            GL.Clear(ClearBufferMask.ColorBufferBit);

            Begin2D();
            
            Blend(BlendState.Normal);

            //Src.One / Dest.OneMinusSrcAlpha for premultiplied textures
            //Src.One / Dest.DstAlpha for additive blending
            //Src.DstColor / Dest.OneMinusSrcAlpha multiply-blending

            /*
            GL.BindTexture(TextureTarget.Texture2D, texture);
			
			GL.Rotate(RotAmount, 0.0, 0.0, 1.0);
			
            GL.Begin(BeginMode.Quads);

            //Goin' clockwise, yo.
            GL.TexCoord2(0.0f, 0.0f); GL.Vertex3(PixToHoriz(91), PixToVert(91), 0.0);
            GL.TexCoord2(1.0f, 0.0f); GL.Vertex3(PixToHoriz(182), PixToVert(91), 0.0);
            GL.TexCoord2(1.0f, 1.0f); GL.Vertex3(PixToHoriz(182), PixToVert(182), 0.0);
            GL.TexCoord2(0.0f, 1.0f); GL.Vertex3(PixToHoriz(91), PixToVert(182), 0.0);

            GL.End();
            */

            for (int X = xMin; X < xMax; X++)
            {
                for (int F = fMin; F < fMax; F++)
                {
                    DrawImage(texSquareGuy, (F * 80), X * 40, 0, 0, 91, 91, 
                        RotAmount);
                }
            }

            SwapBuffers();
        }

        protected void Blend(BlendState blendState)
        {
            GL.BlendFunc(blendState.Source, blendState.Dest);
        }

        #endregion

        #region DrawImage
        public void DrawImage(GLTexture2D texture, double X, double Y, 
            double OrigX, double OrigY,
            int Width, int Height, double RotAmount)
        {
            Vector2d Origin = new Vector2d(OrigX, OrigY);

            Vector2d[] Points = GetCorners(Origin, Width, Height);
            
            GL.LoadIdentity();

            GL.Translate(X, Y, 0);
            GL.BindTexture(TextureTarget.Texture2D, texture.TexID);
            GL.Rotate(RotAmount, 0.0, 0.0, 1.0);

            GL.Begin(BeginMode.Quads);


                #if COLORTEST
                                GL.Color3(1.0, 0.8, 0.8);
                #endif
                            GL.TexCoord2(0.0, 0.0);
                            GL.Vertex2(Points[0]);
                #if COLORTEST
                                GL.Color3(1.0, 1.0, 0.8);
                #endif
                            GL.TexCoord2(texture.ProportionalWidth, 0.0);
                            GL.Vertex2(Points[1]);
                #if COLORTEST
                                GL.Color3(0.8, 1.0, 0.8);
                #endif
                            GL.TexCoord2(texture.ProportionalWidth, 
                                texture.ProportionalHeight);
                            GL.Vertex2(Points[2]);
                #if COLORTEST
                                GL.Color3(0.8, 0.8, 1.0);
                #endif
                            GL.TexCoord2(0.0, texture.ProportionalHeight);
                            GL.Vertex2(Points[3]);

            GL.End();
        }

        private Vector2d[] GetCorners(Vector2d Origin, double Width, double Height)
        {
            
            Vector2d[] Points = new Vector2d[4];
            double CornerX = -Origin.X;
            double CornerY = -Origin.Y;
            double CornerXW = Width - Origin.X;
            double CornerYH = Height - Origin.Y;

            Points[0] = new Vector2d(CornerX, CornerY);
            Points[1] = new Vector2d(CornerXW, CornerY);
            Points[2] = new Vector2d(CornerXW, CornerYH);
            Points[3] = new Vector2d(CornerX, CornerYH);
            
            return Points;
        }
        #endregion


        #region AudioThing
        int[] SoundBuffer = new int[2];
        int SoundSource;
        AudioContext AC;

        bool audioStopped = false;
        bool enterPressed = false;

        private void AudioRun()
        {
            AC = new AudioContext();

            //Generate sound-buffer: testloop beginning
            SoundBuffer[0] = AL.GenBuffer();
            AudioReader sound = new AudioReader("loopytesty b.wav");
            AL.BufferData(SoundBuffer[0], sound.ReadToEnd());

            //Generate sound-buffer: testloop loop
            SoundBuffer[1] = AL.GenBuffer();
            sound = new AudioReader("loopytesty l.wav");
            AL.BufferData(SoundBuffer[1], sound.ReadToEnd());

            SoundSource = AL.GenSource();

            AL.SourceQueueBuffer(SoundSource, SoundBuffer[0]);

            AL.SourcePlay(SoundSource);
        }

        private void AudioEnd()
        {
            AL.SourceStop(SoundSource);
            AL.DeleteSource(SoundSource);

            AL.DeleteBuffers(SoundBuffer);

            AC.Dispose();
        }
        #endregion
        #region public static void Main()

        /// <summary>
        /// Entry point of this example.
        /// </summary>
        [STAThread]
        public static void Main()
        {
            using (Textures example = new Textures())
            {
                // Get the title and category  of this example using reflection.

                example.Run(60.0, 60);
            }
        }

        #endregion
    }
}
